import pdb
import tiktoken
from utils import get_chat_completion

system = """You will be provided with some telemetry data and an issue statement explaining a root cause analysis problem to resolve.

{info}

{data}"""

user = """Now, I need you to provide an root cause analysis to the following question:

```issue
{objective}
```

How to solve a root cause analysis issue:

1. **Follow the workflow of `preprocess -> anomaly detection -> fault identification -> root cause localization` for failure diagnosis.** 
    1.1. Preprocess:
        - Aggregate each KPI of each components that are possible to be the root cause component to obtain multiple time series classified by 'component-KPI' (e.g., service_A-cpu_usage_pct).
        - Then, calculate global thresholds (e.g., global P95, where 'global' means the threshold of all 'component-KPI' time series within a whole metric file) for each 'component-KPI' time series. - Finally, filter data within the given time duration for all time series to perform further analysis.
        - Since the root cause component must be selected from the provided possible root cause components, all other level's components (e.g., service mesh components, middleware components, etc.) should be ignored.
    1.2. Anomaly detection: 
        - An anomaly is typically a data point that exceeds the global threshold.
        - Look for anomalies below a certain threshold (e.g., <=P95, <=P15, or <=P5) in traffic KPIs or business KPIs (e.g., success rate (ss)) since some network failures can cause a sudden drop on them due to packet loss.
        - Loose the global threshold (e.g., from >=P95 to >=P90, or from <=P95 to <=P15, <=P5) if you really cannot find any anomalies.
    1.3. Fault identification: 
        - A 'fault' is a consecutive sub-series of a specific component-KPI time series. Thus, fault identification is the process of identifying which components experienced faults, on which resources, and at what occurrence time points.
        - Filter out isolated noise spikes to locate faults.
        - Faults where the maximum (or minimum) value in the sub-series only slightly exceeds (or falls below) the threshold (e.g., threshold breach <= 50% of the extremal), it’s likely a false positive caused by random KPI fluctuations, and should be excluded.
    1.4. Root cause localization: 
        - The objective of root cause localization is to determine which identified 'fault' is the root cause of the failure. The root cause occurrence time, component, and reason can be derived from the first piece of data point of that fault.
        - If multiple faulty components are identified at **different levels** (e.g., some being containers and others nodes), and all of them are potential root cause candidates, while the issue itself describes a **single failure**, the root cause level should be determined by the fault that shows the most significant deviation from the threshold (i.e., >> 50%).
        - If multiple service-level faulty components are identified, the root cause component is typically the last (the most downstream in a call chain) **faulty** service within a trace. Use traces to identify the root cause component among multiple faulty services.
        - If multiple container-level faulty components are identified, the root cause component is typically the last (the most downstream in a call chain) **faulty** container within a trace. Use traces to identify the root cause component among multiple faulty container.
        - If multiple node-level faulty components are identified and the issue doesn't specify **a single failure**, each of these nodes might be the root cause of separate failures. Otherwise, the predominant nodes with the most faults is the root cause component. The node-level failure do not propagate, and trace only captures communication between all containers or all services.
        - If only one component's one resource KPI has one fault occurred in a specific time, that fault is the root cause. Otherwise, you should use traces and logs to identify the root cause component and reason.
2. **Follow the order of `threshold calculation -> data extraction -> metric analyis -> trace analysis -> log analysis` for failure diagnosis.** 
    2.0. Before analysis: You should extract and filter the data to include those within the failure duration only after the global threshold has been calculated. After these two steps, you can perform metric analysis, trace analysis, and log analysis.
    2.1. Metric analysis: Use metrics to calculate whether each KPIs of each component has consecutive anomalies beyond the global threshold is the fastest way to find the faults. Since there are a large number of traces and logs, metrics analysis should first be used to narrow down the search space of duration and components.
    2.2. Trace analysis: Use traces can further localize which container-level or service-level faulty component is the root cause components when there are multiple faulty components at the same level (container or service) identified by metrics analysis.
    2.3. Log analysis: Use logs can further localize which resource is the root cause reason when there are multiple faulty resource KPIs of a component identified by metrics analysis. Logs can also help to identify the root cause component among multiple faulty components at the same level.
    
Your response should be structured into a JSON format, itemising the relevant root cause information you find. You only need to provide the elements asked by the issue, and ommited the other fields in the JSON. The overall format is as follows:

{{
    "1": {{
        "root cause occurrence datetime": (if asked by the issue, format: '%Y-%m-%d %H:%M:%S', otherwise ommited),
        "root cause component": (if asked by the issue, one selected from the possible root cause component list, otherwise ommited),
        "root cause reason": (if asked by the issue, one selected from the possible root cause reason list, otherwise ommited),
    }}, (mandatory)
    "2": {{
        "root cause occurrence datetime": (if asked by the issue, format: '%Y-%m-%d %H:%M:%S', otherwise ommited),
        "root cause component": (if asked by the issue, one selected from the possible root cause component list, otherwise ommited),
        "root cause reason": (if asked by the issue, one selected from the possible root cause reason list, otherwise ommited),
    }}, (only if the failure number is "unknown" or "more than one" in the issue)
    ... (only if the failure number is "unknown" or "more than one" in the issue)
}}
(DO NOT contain "```json" and "```" tags. DO contain the JSON object with the brackets "{{}}" only.)

Please follow the format above to provide your response of current issue.

Response below:"""

class DirectLM:
    def __init__(self, oracle, schema) -> None:
        self.tokenizer = tiktoken.encoding_for_model("gpt-4o")
        self.oracle = oracle
        self.schema = schema
    
        
    def run(self, instruction, period_data, sample_interval, logger, max_try=3):
        logger.info(f"Objective: {instruction}")
        
        data = f"""## TELEMETRY DATA (Sampled every {sample_interval/60} min):"""
        for key in sorted(period_data.keys()):
            value = period_data[key]
            data += "\n\n" + "".join([f"### {str(key).upper()} DATA", value])
            logger.debug(f"{str(key).upper()} DATA tokens: {len(self.tokenizer.encode(value))}")
        info = self.schema
        prompt = [
                    {'role': 'system', 'content': system.format(info=info, data=data)},
                    {'role': 'user', 'content': user.format(objective=instruction)}
        ]

        logger.debug(f"prompt tokens: {len(self.tokenizer.encode(prompt[0]['content']))}")
        
        for i in range(max_try):
            try:
                response = get_chat_completion(
                    messages=prompt,
                )
                logger.debug(f"Raw Response:\n{response}")
                return response, prompt
            except Exception as e:
                logger.error(e)
                if 'context_length_exceeded' in str(e):
                    logger.error("Token length exceeds the limit.")
                    return "EXCEED!", prompt
        logger.warning("Max steps reached. Please check the history.")
        return "Max steps reached. Please check the history.", prompt
                
                
                