from collections import Counter
import pandas as pd
import numpy as np
import time
import json
import sys
import os
import re

parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
sys.path.insert(0, parent_dir)
from utils import get_chat_completion


system = """Let's play a game. In this game, your task is to check whether the response from the DevOps assistant system meets each scoring criterion. You will assign points to the response based on the given scoring criteria. Your goal is to provide a well-reasoned score that would be convincing to even the most experienced human experts. They should not be able to tell that this score was generated by an AI assistant following specific guidelines.

The information provided to you includes the following components:

```issue
(Instruction for the DevOps assistant system)
```

```prediction
(Response from the DevOps assistant system)
```

```scoring_points
(Scoring criteria for the response)
```

Your response should follow the JSON format below:

{
    "passed criteria": (A list of criteria that the response meets.),
    "failed criteria": (A list of criteria that the response does not meet.),
    "score": (The total score.)
}
(DO NOT contain "```json" and "```" tags. DO contain the JSON object with the brackets "{}" only.)

**RULES OF SCORING POINTS**:

1. A prediction is considered correct if the predicted root cause occurrence datetime is within 1 minutes (i.e., <=1min) of the reference root cause occurrence datetime. For example, if the reference datetime is `2022-03-20 09:09:06` and the predicted time is `2022-03-20 09:09:00`, the prediction is correct.

2. A prediction is incorrect (score=0.0) if it states there is no fault or does not provide information about the failure root cause, when the reference indicates a failure exists. For example, if the prediction is “I cannot find any failure or root cause,” it is incorrect, and you should assign a score of 0.0, regardless of other criteria.

Now, let's get started!"""

user = """Please evaluate the response from the DevOps assistant system based on the scoring points provided. Assign points to the response based on the scoring criteria. If the response meets the criteria, assign the corresponding points. If the response does not meet the criteria, assign 0 points:

```issue
{issue}
```

```prediction
{prediction}
```

```scoring_points
{scoring_points}
```"""

def evaluate_data(reference, predicted):
    from nltk.translate.bleu_score import sentence_bleu, SmoothingFunction
    from rouge_score import rouge_scorer
    import bert_score
    # BLEU
    reference_tokens = [reference.split()]
    generated_tokens = predicted.split()
    smoothing_function = SmoothingFunction().method1
    bleu1_score = sentence_bleu(reference_tokens, generated_tokens, weights=(1, 0, 0, 0), smoothing_function=smoothing_function)
    bleu4_score = sentence_bleu(reference_tokens, generated_tokens, weights=(0.25, 0.25, 0.25, 0.25), smoothing_function=smoothing_function)
    
    # ROUGE
    scorer = rouge_scorer.RougeScorer(['rouge1', 'rougeL'], use_stemmer=True)
    rouge_scores = scorer.score(reference, predicted)
    rouge1_score = rouge_scores['rouge1'].fmeasure
    rougeL_score = rouge_scores['rougeL'].fmeasure
    
    # BERTScore
    P, R, F1 = bert_score.score([predicted], [reference], lang="en")
    
    return {
        'bleu1_score': bleu1_score,
        'bleu4_score': bleu4_score,
        'rouge1': rouge1_score,
        'rougeL': rougeL_score,
        'bert_score_precision': P.mean().item(),
        'bert_score_recall': R.mean().item(),
        'bert_score_f1': F1.mean().item()
    }


def gpt_evaluate_data(dataset: str, issue:str, prediction:str, scoring_points:str, logger=None):
    """Evaluate prediction with GPT

    Args:
        dataset (str): the dataset
        issue (str): question for the DevOps assistant system
        prediction (str): response from the DevOps assistant system
        scoring_points (str): scoring criteria for the response
        
    Given the type of task, load the corresponding scoring points
    Then ask GPT the grant the points for the prediction based on the reference
    """
    
    prompt = [
        {'role': 'system', 'content': system},
        {'role': 'user', 'content': user.format(issue=issue, 
                                                prediction=prediction, 
                                                scoring_points=scoring_points)},
        ]  

    passed_list = []
    failed_list = []
    score_list = [] 
    final_score = None
    response = ""
    for evaluator in range(5):
        for i in range(3): 
            try:
                response = get_chat_completion(
                    messages = prompt,
                    ground=True
                )
                logger.debug(response)
                response = json.loads(response)
                passed_criteria = response['passed criteria']
                failed_criteria = response['failed criteria']
                score = response['score']
                norm_score = float(score) / (len(passed_criteria) + len(failed_criteria))
                score_list.append(norm_score)
                passed_list.append(passed_criteria)
                failed_list.append(failed_criteria)
                break
            except Exception as e:
                if logger != None:
                    logger.error(e)
                time.sleep(1)
    counter = Counter(score_list)
    final_score, frequency = counter.most_common(1)[0]
    final_passed = passed_list[score_list.index(final_score)]
    final_failed = failed_list[score_list.index(final_score)]
    return final_passed, final_failed, final_score

def re_evaluate_data(dataset: str, issue:str, prediction:str, scoring_points:str, logger=None):
    """Evaluate prediction with Regex

    Args:
        dataset (str): the dataset
        issue (str): question for the DevOps assistant system
        prediction (str): response from the DevOps assistant system
        scoring_points (str): scoring criteria for the response
        
    Given the type of task, load the corresponding scoring points
    Then ask Regex the grant the points for the prediction based on the reference
    """

    import itertools

    predict_pattern = (
        r'{\s*'
        r'(?:"root cause occurrence datetime":\s*"(.*?)")?,?\s*'
        r'(?:"root cause component":\s*"(.*?)")?,?\s*'
        r'(?:"root cause reason":\s*"(.*?)")?\s*}'
    )

    predict_matches = re.findall(predict_pattern, prediction)


    predict_results = []
    
    for match in predict_matches:
        datetime_str, component, reason = match
        predict_results.append({
            "root cause occurrence datetime": datetime_str,
            "root cause component": component,
            "root cause reason": reason
        })



    prediction_length = len(predict_results)

    component_pattern = r"The (?:\d+-th|only) predicted root cause component is ([^.]+)"
    reason_pattern = r"The (?:\d+-th|only) predicted root cause reason is ([^.]+)"
    time_pattern = r"The (?:\d+-th|only) root cause occurrence time is within 1 minutes \(i.e., <=1min\) of (.*?)\s*\(?1 point\)?"

    components = re.findall(component_pattern, scoring_points)
    reasons = re.findall(reason_pattern, scoring_points)
    times = re.findall(time_pattern, scoring_points)

    scoringpoints_length = max(len(components),len(reasons),len(times))
    socres_num = len(components)+len(reasons)+len(times)

    def time_difference(time1_str,time2_str):
        from datetime import datetime
        time_format = "%Y-%m-%d %H:%M:%S"
        
        try:
            time1 = datetime.strptime(time1_str, time_format)
            time2 = datetime.strptime(time2_str, time_format)
        except ValueError:
            return False
        
        time_difference = abs(time1 - time2)
        if time_difference.total_seconds() <= 60:
            return True
        else:
            return False

    scores_get = 0
    passing_criteria = []
    failing_criteria = []

    if scoringpoints_length == prediction_length:  
        best_sore = -1
        for perm in itertools.permutations(predict_results):
            current_score = 0
            current_passing = []
            for i in range(scoringpoints_length):
                if len(components) == scoringpoints_length:
                    if perm[i]['root cause component'] == components[i]:
                        current_score +=1
                        current_passing.append(components[i])
                if len(reasons) == scoringpoints_length:
                    if perm[i]['root cause reason'] == reasons[i]:
                        current_score +=1
                        current_passing.append(reasons[i])
                if len(times) == scoringpoints_length:
                    if time_difference(times[i],perm[i]['root cause occurrence datetime']):
                        current_score +=1
                        current_passing.append(times[i])
            if current_score>best_sore:
                best_sore = current_score
                passing_criteria = current_passing
        scores_get = best_sore            
    
    failing_criteria = list(set(components+reasons+times)-set(passing_criteria))
    
    final_score = scores_get/socres_num
    return passing_criteria, failing_criteria, final_score
    

def evaluate_csv(filename, reference_column_name='reference', predicted_column_name='predicted'):
    data = pd.read_csv(filename)
    results = []
    for index, row in data.iterrows():
        reference = row[reference_column_name]
        predicted = row[predicted_column_name]
        scores = evaluate_data(reference, predicted)
        results.append({
            'reference': reference,
            'predicted': predicted,
            'scores': scores
        })
    return results

def save_results(results, filename=None):
    bleu1_scores = [r['scores']['bleu1_score'] for r in results]
    bleu4_scores = [r['scores']['bleu4_score'] for r in results]
    rouge1_scores = [r['scores']['rouge1'] for r in results]
    rougeL_scores = [r['scores']['rougeL'] for r in results]
    bert_precision = [r['scores']['bert_score_precision'] for r in results]
    bert_recall = [r['scores']['bert_score_recall'] for r in results]
    bert_f1 = [r['scores']['bert_score_f1'] for r in results]

    avg_bleu1_scores = np.mean(bleu1_scores)
    avg_bleu4_scores = np.mean(bleu4_scores)
    avg_rouge1_scores = np.mean(rouge1_scores)
    avg_rougeL_scores = np.mean(rougeL_scores)
    avg_bert_precision = np.mean(bert_precision)
    avg_bert_recall = np.mean(bert_recall)
    avg_bert_f1 = np.mean(bert_f1)

    print(f"Average BLEU-1 Score: {avg_bleu1_scores}")
    print(f"Average BLEU-4 Score: {avg_bleu4_scores}")
    print(f"Average ROUGE-1 Score: {avg_rouge1_scores}")
    print(f"Average ROUGE-L Score: {avg_rougeL_scores}")
    print(f"Average BERTScore Precision: {avg_bert_precision}")
    print(f"Average BERTScore Recall: {avg_bert_recall}")
    print(f"Average BERTScore F1: {avg_bert_f1}")

    if filename:
        with open(filename, 'w') as f:
            f.write(f"BLEU-1 Score: {avg_bleu1_scores}\n")
            f.write(f"BLEU-4 Score: {avg_bleu4_scores}\n")
            f.write(f"ROUGE-1 Score: {avg_rouge1_scores}\n")
            f.write(f"ROUGE-L Score: {avg_rougeL_scores}\n")
            f.write(f"BERTScore Precision: {avg_bert_precision}\n")
            f.write(f"BERTScore Recall: {avg_bert_recall}\n")
            f.write(f"BERTScore F1: {avg_bert_f1}\n")

    