import os
import json
import yaml
import time

def load_config(config_path="./api_config.yaml"):
    configs = dict(os.environ)
    with open(config_path, "r") as file:
        yaml_data = yaml.safe_load(file)
    configs.update(yaml_data)
    return configs

configs = load_config()

def OpenAI_chat_completion(model, messages, temperature):    
    from openai import OpenAI
    os.environ["OPENAI_API_KEY"] = configs["API_KEY"]
    if "OPENAI_API_BASE" in configs.keys():
        client = OpenAI(base_url=configs["API_BASE"])
    return client.chat.completions.create(
        model = model,
        messages = messages,
        temperature = temperature,
    ).choices[0].message.content

def AI_chat_completion(model, messages, temperature):    
    from openai import OpenAI
    client = OpenAI(
        api_key=configs["API_KEY"],
        base_url=configs["API_BASE"]
    )
    return client.chat.completions.create(
        model = model,
        messages = messages,
        temperature = temperature,
    ).choices[0].message.content

def Google_chat_completion(model, messages, temperature):
    import google.generativeai as genai
    genai.configure(api_key=configs["API_KEY"])
    genai.GenerationConfig(temperature=temperature)
    system_instruction = messages[0]["content"] if messages[0]["role"] == "system" else None
    messages = [item for item in messages if item["role"] != "system"]
    messages = [{"role": "model" if item["role"] == "assistant" else item["role"], "parts": item["content"]} for item in messages]
    history = messages[:-1]
    message = messages[-1]
    chat = genai.GenerativeModel(
        model_name=model,
        system_instruction=system_instruction
        ).start_chat(history=history if history != [] else None)
    return chat.send_message(message).text


def get_chat_completion(messages, ground=False, temperature=0.0):
    """return the completion from Azure or OpenAI or other models
    
    Args:
        messages (Iterable[ChatCompletionMessageParam]): The chat message for completion
    
    Return: ChatCompletion 
    """
    if ground:
        model = configs['GROUND_MODEL']
    else:
        model = configs["MODEL"]

    def send_request():
        if configs["SOURCE"] == "AI":
            return AI_chat_completion(model, messages, temperature)
        elif configs["SOURCE"] == "OpenAI":
            return AI_chat_completion(model, messages, temperature)
        elif configs["SOURCE"] == "Google":
            return Google_chat_completion(model, messages, temperature)
    
    for i in range(3):
        try:
            return send_request()
        except Exception as e:
            print(e)
            if '429' in str(e):
                print("Rate limit exceeded. Waiting for 1 second.")
                time.sleep(1)
                continue
            else:
                raise e
    
    
def load_task_templates(fpath="task_templates.json")->dict:
    """_summary_

    Args:
        fpath (str, optional): Defaults to "task_templates.json".

    Returns:
        dict: A dict that stores the templates or each task, with 'insruction', 'suffix',
        'response' and 'scoring_points' as keys of each item.
    """
    
    with open(fpath, "r", encoding="utf8") as f:
        task_templates = json.load(f)
    return task_templates